# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import inspect
import locale
import os
import time
from copy import deepcopy
import traceback

import sys
import requests
from requests.exceptions import ReadTimeout
from . import settings, global_cache_map

from ..uqer import session
from ..config import *
from ..version import __version__

from platform import python_version
import json
import codecs

csv_encoding = 'gb2312'
request_encoding = 'UTF-8'

timeout = 60
retry_interval = 2
max_retries = 5

client_info = json.dumps({"python_version": python_version(),
"client_version": __version__,
"module":"uqer_sdk"})

def get_real_string(input_string):
    if sys.version_info.major == 2:
        out = input_string.encode(locale.getpreferredencoding())
    else:
        out = input_string
    return out


def get_http_result(httpClient, requestString, gw, max_retries=max_retries):
 
    for i in range(1, max_retries + 1):
        try:
            source = 'mercury_sdk:' + os.environ.get('DatayesPrincipalName', 'invalid_user')
            result = session.get("https://%s:%d/data/v1%s" % (httpClient[2], httpClient[3], requestString),
                                  headers={'Connection': 'close', 
                                           "Authorization": "Bearer " + get_token(), 
                                           'SOURCE': source,
                                           'CLIENT_INFO': client_info},
                                  timeout=timeout)
            return result

        except Exception as e:
            time.sleep(retry_interval)


def __getCSV__(requestString, httpClient, gw=True):
    try:
        result = get_http_result(httpClient, requestString, gw)
        if result.status_code == 400:
            raise Exception('请检查输入参数，可能某列表输入参数过长')
        if int(result.headers.get('dyes-rsp-count', 0)) == 100000:
            result2 = get_http_result(httpClient, requestString + '&pagenum=2', gw)
            if int(result2.headers.get('dyes-rsp-count', 0)) > 0:
                raise Exception(get_real_string('Data API查询返回结果超过数据记录上限（十万条记录），请修改查询条件分步查询'))
        return result.text
    except ReadTimeout:
        raise Exception('查询服务超时')
    except Exception as e:
        raise e

def get_cache_key(frame):
    args, _, _, values = inspect.getargvalues(frame)
    func_name = inspect.getframeinfo(frame)[2]
    cache_key = hash([values[arg] for arg in args].__str__())
    return func_name, cache_key

def get_data_from_cache(func_name, cache_key):
    return

def put_data_in_cache(func_name, cache_key, data):
    return

def splist(l, s):
    return [l[i:i+s] for i in range(len(l)) if i %s == 0]

def is_no_data_warn(csvString, print_msg):
    if csvString.startswith('-1:No Data Returned'):
        if print_msg:
            print('没有数据返回。请检查输入参数，若仍有问题，可联系service.uqer@datayes.com')
        return True
    return False

def handle_error(csvString, api_name):
    if csvString.startswith('-403:Need Privilege'):
        result = '无%s接口使用权限，您可以购买优矿专业版（https://uqer.io/pro） 或联系 4000 820 386 购买数据' % api_name
    elif csvString.startswith('-403:Need login'):
        result = '您未登陆'
    elif csvString.startswith('-2:Invalid Request Parameter'):
        result = '无效的请求参数。请检查输入参数，若仍有问题，可联系service.uqer@datayes.com'
    elif csvString.startswith('-3:Service Suspend'):
        result = '服务终止。请联系service.uqer@datayes.com'
    elif csvString.startswith('-4:Internal Server Error'):
        result = '内部服务器错误。请联系service.uqer@datayes.com'
    elif csvString.startswith('-5:Server Busy'):
        result = '服务器拥堵。可能是海量用户在同一时间集中调用该数据造成，可稍后再次尝试。' \
                 '如长时间未改善，或频繁出现该问题，可联系sevice.uqer@datayes.com'
    elif csvString.startswith('-6:Trial Times Over'):
        result = '试用次数达到限制。您对该数据的试用权限已经到期'
    elif csvString.startswith('-7:Query Timeout'):
        result = '请求超时。可能您请求的数据量较大或服务器当前忙'
    elif csvString.startswith('-8:Query Failed'):
        result = '请求失败，请联系service.uqer@datayes.com'
    elif csvString.startswith('-9:Required Parameter Missing'):
        result = '必填参数缺失。请仔细复核代码，将其中的参数补充完整后再次尝试'
    elif csvString.startswith('-11:The number of API calls reached limit'):
        result = '当日调用次数达到上限，请优化代码调用逻辑。每日0点重新计数'
    else:
        result = csvString
    err_msg = result
    print (err_msg)

    raise Exception(get_real_string(err_msg))

def get_token():
    import os
    ACCESS_TOKEN = "cloud_sso_token"
    DataAPI_TOKEN = "access_token"

    if os.environ.get(DataAPI_TOKEN):
        return os.environ[DataAPI_TOKEN]

    if os.environ.get(ACCESS_TOKEN):
        access_token = os.environ[ACCESS_TOKEN]
        r2 = None
        try:
            r2 = session.post(TOKEN_URL,
                               data={'grant_type': 'permanent'})
            token = r2.json()['access_token']
            os.environ[DataAPI_TOKEN] = token

            return token
        except:
            print ('获取通联数据权限凭证出错，可能是您的token无效或者token过期，请重新登录\n')

            return ''
    else:
        print ('当前环境下没有 token，可能是您未登录，请先登录哦')
        return ''

def __getConn__():
    return server

def __formatDate__(inputDate):
    return inputDate

def lowcase_keys(d):
    result = {}
    for key, value in d.items():
        lower_key = key.lower()
        result[lower_key] = value
    return result

def is_pro_user():
    return True


def showtraceback(self, exc_tuple=None, filename=None, tb_offset=None,
                  exception_only=False):
    import traceback
    import sys

    etype, value, tb = self._get_exc_info(exc_tuple)
    listing = traceback.format_exception(etype, value, tb)
    last_message = ''
    lineno = None
    text = ''

    if listing:
        last_message = listing[-1].decode('utf-8')
    
    except_msg = 'Exception:'
    if last_message.startswith(except_msg):
        last_message = '异常:' + last_message[len(except_msg):]

    for filename, lineno, module, text in traceback.extract_tb(tb):
        if filename.startswith('<mercury-input') and \
           not text.startswith('display=True, return_quartz_data=True'):
            break

    if lineno:
        print('行号: %s\n代码: %s\n%s' %(lineno, text, last_message))
    else:
        print(last_message)


def pretty_traceback():
    if not settings.pretty_traceback_enabled:
        return
    import os
    if os.environ.get('env') in ['qa', 'stg', 'prd']:
        try:
            import IPython
            IPython.core.interactiveshell.InteractiveShell.showtraceback = showtraceback
        except:
            pass


from .version import __version__ as dataapi_version

version_url = '%s/w/api/info'%(SDK_URL)
files_url = '%s/w/api/download'%(SDK_URL)


def get_api_version():
    res = requests.get(version_url).json()
    if res.get('code') != 200:
        return '', []

    remote_version = res['data']['version']
    py_files = res['data']['py_files']

    return remote_version, py_files


def get_api_file(filename):
    code = 0
    text = ''

    for i in range(3):
        try:
            response = requests.get('%s/%s'%(files_url, filename))
            code, text = response.status_code, response.text
            if code == 200:
                break
        except:
            pass
        import time;time.sleep(1)
    
    return code, text

def replace_api_files():
    dir_path = os.path.dirname(os.path.realpath(__file__))
    remote_version, py_files = get_api_version()
    if not remote_version or remote_version <= dataapi_version:
        return

    for filename in py_files:
        if filename == '__init__':
            continue

        filename = '%s.py'%(filename)
        file_path = os.path.join(dir_path, filename)

        status_code, response = get_api_file(filename)

        if status_code!=200:
            # 文件可能是没下载正确，就不更新了
            continue

        if os.path.isfile(file_path):
            os.remove(file_path)

        with codecs.open(file_path, 'w', "utf-8-sig") as f:
            f.write(response)

    version_str = "__version__ = '%s'"%(remote_version)
    version_file_path = os.path.join(dir_path, 'version.py')
    with open(version_file_path, 'w') as f:
        f.write(version_str)
        print('UQER SDK的DataAPI模块版本由%s升级到%s'%(dataapi_version, remote_version))


# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import requests
import os

from .config import *
from .version import __version__ as v

version = 'sdk_%s'%v

class Notebook(object):

    def __init__(self, session, target_path):
        self.session = session

        if target_path:
            self.target_path = target_path
        else:
            self.target_path = os.getcwd()

        self.notebook = self.list()


    def list(self):
        self.notebook = self._list_notebook()
        return self.notebook


    def _list_notebook(self):
        url = NOTEBOOK_URL
        res = self.session.get(url)
        if not res.ok:
            return 0
        data = res.json()

        if isinstance(data, dict) and data.get('code'):
            print(data)
            return []

        all_notebook = []
        self.directories = []

        for elem in data:
            name = elem.get('name')
            type = elem.get('type')

            if type == 'notebook':
                all_notebook.append(name)

            elif type == 'directory':
                self.directories.append(name)
                sub_elems = elem.get('children', [])
                for sub_elem in sub_elems:
                    elem_name = sub_elem.get('name')
                    type = sub_elem.get('type')
                    if type == 'notebook':
                        all_notebook.append('%s/%s'%(name, elem_name))

        return all_notebook
        

    def backup(self):
        print("该方法已废弃，请使用 client.notebook.download(download_all=True) 下载所有的notebook文件。")


    def download(self, filename='', target_path='', download_all=False):

        if target_path:
            self.target_path = target_path  

        self.notebook = self.list()
        directories = self.directories

        if download_all:
            for i in self.notebook:
                self._download_notebook(i)

        elif type(filename) == list:
            for i in filename:
                # 以ipynb格式请求后端文件
                i = i.replace('.nb', '.ipynb')
                if i in self.notebook:
                    self._download_notebook(i)
                elif i in directories:
                    self._download_dir(i)
                else:
                    print ('文件不存在 {}'.format(i))

        elif type(filename) in (str, unicode):
            # 以ipynb格式请求后端文件
            filename = filename.replace('.nb', '.ipynb')
            if filename in self.notebook:
                self._download_notebook(filename)
            elif filename in directories:
                self._download_dir(filename)
            else:
                print ('文件不存在 {}'.format(filename))
        else:
            pass


    def _download_dir(self, dir):
        dir =dir.replace('/', '')
        dir_elems = [e for e in self.notebook if e.startswith('%s/'%dir)]
        for dir_elem in dir_elems:
            self._download_notebook(dir_elem)


    def _download_notebook(self, filename):
        url = DOWN_NOTEBOOK_URL
        
        notebook_url = url + '/' + requests.utils.quote(filename.encode('utf-8'))

        # 以nb格式保存到本地文件
        filename = filename.replace('.ipynb', '.nb')
        print('下载Notebook：\n开始下载 %s' %filename)

        target_path = self.target_path

        file_path = os.path.join(target_path, filename)

        try:
            path, filename = file_path.rsplit('/', 1)
        except:
            path = ''

        if path and not os.path.exists(path):
            os.mkdir(path)
            print('文件夹 %s 不存在，已自动创建。'%path)

        with open(file_path, 'wb') as f:
            response = self.session.get('%s?%s'%(notebook_url, version), stream=True)

            if not response.ok:
                if response.status_code == 413:
                    print('{}文件超过100兆，暂不支持下载。'.format(filename))
                else:
                    print ('下载文件 {} 过程中出错'.format(filename))
                return 0
            
            for chunk in response.iter_content(1024 * 100):
                f.write(chunk)

        print('完成下载 %s, 路径：%s' %(filename, file_path))


    def delete(self, filename):
        try:
            _delete_notebook(filename)
            print ('删除成功 {}'.format(filename))
        except:
            print ('删除过程出错：', traceback.print_exc())

    def _delete_notebook(self, filename):
        url = NOTEBOOK_DELETE()
        dataurl = url.format(filename)
        self.session.delete('%s?%s'%(dataurl, version))

        print('%s 数据文件删除成功' %(filename))









# -*- coding: utf-8 -*-
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import requests
import traceback
import os
import json
import pandas as pd
import numpy as np
import sys
from . signal_process_check import *
from ..uqer import session

from ..config import *


def neutralize(raw_data, target_date, industry_type="SW1", exclude_style_list=[], **kwargs):
    '''
    中性化处理函数

    :param raw_data: [Dict/Series] 输入待处理因子，key为ticker/secID，value为因子值
    :param target_date: [String] 中性化的基准日期，合理输入为调仓前一个交易日，日期格式为'YYYYMMDD'
    :param industry_type: [String] 行业分类选择, 包括申万一级、申万二级, 分别用SW1/SW2表示，默认为SW1
    :param exclude_style_list: [List] 不中性的风格因子，列入其中的因子在做中性化时不考虑，可选值为'BETA', 'RESVOL', 'MOMENTUM',
    'SIZE', 'SIZENL', 'EARNYILD', 'BTOP', 'GROWTH', 'LEVERAGE', 'LIQUIDTY'，分别代表贝塔、残差波动率、动量、市值、非线性市值、
    盈利性、净市率、成长性、杠杆和流动性
    :return: [Dict/Series] 经过中性化处理之后的因子值
    '''
    # Check Input
    raw_data_check(raw_data)
    industry_check(industry_type)
    date_check(target_date)
    if exclude_style_list:
        risk_factor_check(exclude_style_list)

    if len(raw_data) < 11:
        raise InputLengthError('Neutralize input length too short')

    # 遗留参数处理
    if 'risk_module' in kwargs:
        print ("risk_module参数已经不再使用，中性化函数使用通联风险模型")
    # Series to Dict(if raw_data is Series)
    flag = True
    if isinstance(raw_data, pd.Series):
        flag = False
        raw_data = raw_data.to_dict()

    # Ticker to SecID
    isTicker = False
    for k, v in raw_data.items():
        if len(k) == 11:
            isTicker = True
            break
    record_ticker = {}
    if isTicker:
        for k, v in raw_data.items():
            record_ticker[k[:6]] = k
    # Deal with Nan
    record_nan = {}
    input_dict = {}
    for k, v in raw_data.items():
        if isinstance(v, (int, float)) and math.isnan(v):
            record_nan[k[:6]] = v
        elif v == np.inf or v == -np.inf:
            record_nan[k[:6]] = v
        elif isinstance(v, np.generic):
            input_dict[k[:6]] = np.asscalar(v)
        else:
            input_dict[k[:6]] = v

    send_data = {}
    config_dict = {}
    config_dict['date'] = target_date.replace('-', '')
    config_dict['industry_type'] = industry_type
    config_dict['exclude'] = exclude_style_list
    send_data['raw_data'] = input_dict
    send_data['config'] = config_dict
    try:
        r = session.post(EXT_MFHANDLER_HOST + '/neutralize', json=send_data)
    except:
        raise MFNetworkError('Neutralize Network Error')
    result = handle_recv_json(r)

    # Deal with Nan, Add Nan to result
    if record_nan:
        for k, v in record_nan.items():
            result[k] = v

    # Ticker to SecID, change the key of dict
    final_result = {}
    if isTicker:
        for k, v in result.items():
            final_result[record_ticker[k]] = v
    else:
        final_result = result

    # If Series
    if not flag:
        return pd.Series(final_result)
    else:
        return final_result


# @exception_decorator
def standardize(raw_data):
    '''
    正态化处理函数

    :param raw_data: [Dict/Series] 输入待处理因子，key为ticker/secID，value为因子值
    :return: [Dict/Series] 经过正态化处理之后的因子值
    '''
    raw_data_check(raw_data)
    is_series = False
    if isinstance(raw_data, pd.Series):
        is_series = True

    # Local Process
    se_signal = pd.Series(raw_data)
    if isnull(se_signal):
        return se_signal.to_dict()
    std = se_signal.std()
    if std == 0:
        if is_series:
            return se_signal - se_signal[0]
        else:
            return (se_signal - se_signal[0]).to_dict()  # an all-zero Series
    else:
        if is_series:
            return ((se_signal - se_signal.mean()) / float(std))
        else:
            return ((se_signal - se_signal.mean()) / float(std)).to_dict()


# @exception_decorator
def normalize_l1(raw_data):
    '''
    L1范数正则化函数

    :param raw_data: [Dict/Series] 输入待处理因子，key为ticker/secID，value为因子值
    :return: [Dict/Series] 经过L1范数正则化处理之后的因子值
    '''
    is_series = False
    if isinstance(raw_data, pd.Series):
        is_series = True
    raw_data_check(raw_data)

    # Local Process
    se_signal = pd.Series(raw_data)
    if isnull(se_signal):
        return se_signal.to_dict()
    se_signal = se_signal - se_signal.mean()
    sum_sig = se_signal.abs().sum()
    if sum_sig == 0:
        if is_series:
            return se_signal
        else:
            return se_signal.to_dict()
    else:
        if is_series:
            return se_signal / sum_sig
        else:
            return (se_signal / sum_sig).to_dict()


# @exception_decorator
def winsorize(raw_data, win_type='NormDistDraw', n_draw=5, pvalue=0.05):
    '''
    极值处理函数

    :param raw_data: [Dict/Series] 输入待处理因子，key为ticker/secID，value为因子值
    :param win_type: [String] 去极值处理的类型选择, 包括正态分布去极值和分位数去极值，分别为'NormDistDraw'/'QuantileDraw', 默认为前者
    :param n_draw: [int] 正态分布去极值的迭代次数，只有当win_type='NormDistDraw'，更改该参数才有意义；合法输入为正整数，默认值为5
    :param pvalue: [float] 分位数去极值的分位数指定，只有当win_type='QuantileDraw'，更改该参数才有意义；合法输入为(0,1)区间内的浮点数，默认值为0.05
    :return: [Dict/Series] 经过去极值处理之后的因子值
    '''

    is_series = False
    if isinstance(raw_data, pd.Series):
        is_series = True

    raw_data_check(raw_data)
    if win_type not in ['NormDistDraw', 'QuantileDraw']:
        raise InputConfigError('win_type config error')

    # Local Process
    se_signal = pd.Series(raw_data)
    if isnull(se_signal):
        return se_signal.to_dict()

    if not isinstance(pvalue, float):
        raise InputTypeError('pvalue type error')
    if not 0 < pvalue < 1:
        raise InputConfigError('pvalue config error')

    if not isinstance(n_draw, int) or n_draw <= 0:
        raise InputConfigError('n_draw config error')
    if 'maxint' in sys.__dict__ and n_draw > sys.maxint:
        raise InputConfigError('n_draw is larger than sys.maxint')

    if win_type == 'QuantileDraw':
        QD = QuantileDraw(pvalue)
        if is_series:
            return QD.winsorize(se_signal)
        else:
            return QD.winsorize(se_signal).to_dict()
    else:
        NDD = NormDistDraw(n_draw)
        if is_series:
            return NDD.winsorize(se_signal)
        else:
            return NDD.winsorize(se_signal).to_dict()


class QuantileDraw:
    def __init__(self, pvalue):
        self.pvalue = pvalue

    def winsorize(self, signal):
        series = signal.copy()  # do not modify input data
        lowerbound = self.pvalue / 2.    # quantile lowerbound
        upperbound = 1 - self.pvalue / 2.  # quantile upperbound
        bott = pd.Series.quantile(series, lowerbound)
        upper = pd.Series.quantile(series, upperbound)
        series[series < bott] = bott
        series[series > upper] = upper
        return series


class NormDistDraw:
    def __init__(self, n_draw):
        self.n_draw = n_draw

    def winsorize(self, signal):
        series = signal.copy()  # do not modify input data
        for i in range(self.n_draw):
            std = series.std()
            mean = series.mean()
            series[series < mean - 3 * std] = mean - 3 * std
            series[series > mean + 3 * std] = mean + 3 * std
        return series



# -*- coding: UTF-8 -*-
from . import api_base
try:
    from StringIO import StringIO
except:
    from io import StringIO
import pandas as pd
import sys
from datetime import datetime
from .api_base import get_cache_key, get_data_from_cache, put_data_in_cache, pretty_traceback
import inspect
try:
    unicode
except:
    unicode = str

__doc__="福汇"
def __FXTickRTSnapshotGet(currencyPair, field = "", pandas = "1"):
    """
    获取福汇外汇市场信息快照
    
    :param currencyPair: 货币对，包含以下几种：USD|CHF，EUR|USD，GBP|USD，GBP|AUD，USD|CAD，NZD|USD，EUR|JPY，NZD|CAD，AUD|NZD，EUR|GBP，GBP|CAD，GBP|JPY，AUD|JPY，USD|JPY，EUR|CHF，USD|CNH，EUR|AUD，AUD|CHF，AUD|USD，USD|HKD，不输入返回全部,可以是列表
    :param field: 所需字段,可以是列表,可空
    :param pandas: 1表示返回 pandas data frame，0表示返回csv,可空
    :return: :raise e: API查询的结果，是CSV或者被转成pandas data frame；若查询API失败，返回空data frame； 若解析失败，则抛出异常
    """
        
    pretty_traceback()
    frame = inspect.currentframe()
    func_name, cache_key = get_cache_key(frame)
    cache_result = get_data_from_cache(func_name, cache_key)
    if cache_result is not None:
        return cache_result
    split_index = None
    split_param = None
    httpClient = api_base.__getConn__()    
    requestString = []
    requestString.append('/api/foreignMarket/getFXTickRTSnapshot.csv?ispandas=1&') 
    requestString.append("currencyPair=")
    if hasattr(currencyPair,'__iter__') and not isinstance(currencyPair, str):
        if len(currencyPair) > 100 and split_param is None:
            split_index = len(requestString)
            split_param = currencyPair
            requestString.append(None)
        else:
            requestString.append(','.join(currencyPair))
    else:
        requestString.append(currencyPair)
    requestString.append("&field=")
    if hasattr(field,'__iter__') and not isinstance(field, str):
        if len(field) > 100 and split_param is None:
            split_index = len(requestString)
            split_param = field
            requestString.append(None)
        else:
            requestString.append(','.join(field))
    else:
        requestString.append(field)
    if split_param is None:
        csvString = api_base.__getCSV__(''.join(requestString), httpClient, gw=True)
        if csvString is None or len(csvString) == 0 or (csvString[0] == '-' and not api_base.is_no_data_warn(csvString, False)) or csvString[0] == '{':
            api_base.handle_error(csvString, '__FXTickRTSnapshotGet')
        elif csvString[:2] == '-1':
            csvString = ''
    else:
        p_list = api_base.splist(split_param, 100)
        csvString = []
        for index, item in enumerate(p_list):
            requestString[split_index] = ','.join(item)
            temp_result = api_base.__getCSV__(''.join(requestString), httpClient, gw=True)
            if temp_result is None or len(temp_result) == 0 or temp_result[0] == '{' or (temp_result[0] == '-' and not api_base.is_no_data_warn(temp_result, False)):
                api_base.handle_error(temp_result, '__FXTickRTSnapshotGet')
            if temp_result[:2] != '-1':
                csvString.append(temp_result if len(csvString) == 0 else temp_result[temp_result.find('\n')+1:])
        csvString = ''.join(csvString)
    
    if len(csvString) == 0:
        if 'field' not in locals() or len(field) == 0:
            field = [u'currencyPair', u'offerID', u'quoteID', u'tradeDate', u'dataTime', u'bid', u'ask', u'Low', u'High', u'volume', u'valueDate', u'buyInterest', u'sellInterest', u'contractCurrency', u'contractMultiplier', u'Digits', u'pointSize', u'tradingStatus', u'localTimestamp']
        if hasattr(field, '__iter__') and not isinstance(field, str):
            csvString = ','.join(field) + '\n'
        else:
            csvString = field + '\n'
    if pandas != "1":
        put_data_in_cache(func_name, cache_key, csvString)
        return csvString
    try:
        myIO = StringIO(csvString)
        pdFrame = pd.read_csv(myIO, dtype = {'currencyPair': 'str','quoteID': 'str','tradeDate': 'str','dataTime': 'str','valueDate': 'str','contractCurrency': 'str','tradingStatus': 'str'},  )
        put_data_in_cache(func_name, cache_key, pdFrame)
        return pdFrame
    except Exception as e:
        raise e
    finally:
        myIO.close()

def __FXTickRTIntraDayGet(currencyPair, field = "", pandas = "1"):
    """
    获取福汇外汇市场当日Tick数据信息
    
    :param currencyPair: 货币对，包含EUR|USD、USD|JPY、GBP|USD、USD|CHF、EUR|CHF、AUD|USD、AUD|CHF、USD|CAD、NZD|USD、EUR|GBP、USD|CNH、EUR|JPY、GBP|JPY、EUR|AUD、AUD|JPY、USD|HKD、GBP|CAD、GBP|AUD、NZD|CAD、AUD|NZD
    :param field: 所需字段,可以是列表,可空
    :param pandas: 1表示返回 pandas data frame，0表示返回csv,可空
    :return: :raise e: API查询的结果，是CSV或者被转成pandas data frame；若查询API失败，返回空data frame； 若解析失败，则抛出异常
    """
        
    pretty_traceback()
    frame = inspect.currentframe()
    func_name, cache_key = get_cache_key(frame)
    cache_result = get_data_from_cache(func_name, cache_key)
    if cache_result is not None:
        return cache_result
    split_index = None
    split_param = None
    httpClient = api_base.__getConn__()    
    requestString = []
    requestString.append('/api/foreignMarket/getFXTickRTIntraDay.csv?ispandas=1&') 
    if not isinstance(currencyPair, str) and not isinstance(currencyPair, unicode):
        currencyPair = str(currencyPair)

    requestString.append("currencyPair=%s"%(currencyPair))
    requestString.append("&field=")
    if hasattr(field,'__iter__') and not isinstance(field, str):
        if len(field) > 100 and split_param is None:
            split_index = len(requestString)
            split_param = field
            requestString.append(None)
        else:
            requestString.append(','.join(field))
    else:
        requestString.append(field)
    if split_param is None:
        csvString = api_base.__getCSV__(''.join(requestString), httpClient, gw=True)
        if csvString is None or len(csvString) == 0 or (csvString[0] == '-' and not api_base.is_no_data_warn(csvString, False)) or csvString[0] == '{':
            api_base.handle_error(csvString, '__FXTickRTIntraDayGet')
        elif csvString[:2] == '-1':
            csvString = ''
    else:
        p_list = api_base.splist(split_param, 100)
        csvString = []
        for index, item in enumerate(p_list):
            requestString[split_index] = ','.join(item)
            temp_result = api_base.__getCSV__(''.join(requestString), httpClient, gw=True)
            if temp_result is None or len(temp_result) == 0 or temp_result[0] == '{' or (temp_result[0] == '-' and not api_base.is_no_data_warn(temp_result, False)):
                api_base.handle_error(temp_result, '__FXTickRTIntraDayGet')
            if temp_result[:2] != '-1':
                csvString.append(temp_result if len(csvString) == 0 else temp_result[temp_result.find('\n')+1:])
        csvString = ''.join(csvString)
    
    if len(csvString) == 0:
        if 'field' not in locals() or len(field) == 0:
            field = [u'updateDate', u'updateTime', u'offerID', u'currencyPair', u'quoteId', u'bid', u'ask', u'low', u'high', u'volume', u'sellInterest', u'buyInterest', u'contractCurrency', u'digits', u'pointSize', u'contractMultiplier', u'tradingStatus', u'valueDate']
        if hasattr(field, '__iter__') and not isinstance(field, str):
            csvString = ','.join(field) + '\n'
        else:
            csvString = field + '\n'
    if pandas != "1":
        put_data_in_cache(func_name, cache_key, csvString)
        return csvString
    try:
        myIO = StringIO(csvString)
        pdFrame = pd.read_csv(myIO, dtype = {'currencyPair': 'str','quoteId': 'str','contractCurrency': 'str','tradingStatus': 'str','valueDate': 'str'},  )
        put_data_in_cache(func_name, cache_key, pdFrame)
        return pdFrame
    except Exception as e:
        raise e
    finally:
        myIO.close()

def __FXBarRTIntraDayGet(currencyPair, field = "", pandas = "1"):
    """
    高频数据，获取福汇外汇交易市场当日分钟线数据
    
    :param currencyPair: 货币对，包含EUR|USD、USD|JPY、GBP|USD、USD|CHF、EUR|CHF、AUD|USD、AUD|CHF、USD|CAD、NZD|USD、EUR|GBP、USD|CNH、EUR|JPY、GBP|JPY、EUR|AUD、AUD|JPY、USD|HKD、GBP|CAD、GBP|AUD、NZD|CAD、AUD|NZD,可以是列表
    :param field: 所需字段,可以是列表,可空
    :param pandas: 1表示返回 pandas data frame，0表示返回csv,可空
    :return: :raise e: API查询的结果，是CSV或者被转成pandas data frame；若查询API失败，返回空data frame； 若解析失败，则抛出异常
    """
        
    pretty_traceback()
    frame = inspect.currentframe()
    func_name, cache_key = get_cache_key(frame)
    cache_result = get_data_from_cache(func_name, cache_key)
    if cache_result is not None:
        return cache_result
    split_index = None
    split_param = None
    httpClient = api_base.__getConn__()    
    requestString = []
    requestString.append('/api/foreignMarket/getFXBarRTIntraDay.csv?ispandas=1&') 
    requestString.append("currencyPair=")
    if hasattr(currencyPair,'__iter__') and not isinstance(currencyPair, str):
        if len(currencyPair) > 100 and split_param is None:
            split_index = len(requestString)
            split_param = currencyPair
            requestString.append(None)
        else:
            requestString.append(','.join(currencyPair))
    else:
        requestString.append(currencyPair)
    requestString.append("&field=")
    if hasattr(field,'__iter__') and not isinstance(field, str):
        if len(field) > 100 and split_param is None:
            split_index = len(requestString)
            split_param = field
            requestString.append(None)
        else:
            requestString.append(','.join(field))
    else:
        requestString.append(field)
    if split_param is None:
        csvString = api_base.__getCSV__(''.join(requestString), httpClient, gw=True)
        if csvString is None or len(csvString) == 0 or (csvString[0] == '-' and not api_base.is_no_data_warn(csvString, False)) or csvString[0] == '{':
            api_base.handle_error(csvString, '__FXBarRTIntraDayGet')
        elif csvString[:2] == '-1':
            csvString = ''
    else:
        p_list = api_base.splist(split_param, 100)
        csvString = []
        for index, item in enumerate(p_list):
            requestString[split_index] = ','.join(item)
            temp_result = api_base.__getCSV__(''.join(requestString), httpClient, gw=True)
            if temp_result is None or len(temp_result) == 0 or temp_result[0] == '{' or (temp_result[0] == '-' and not api_base.is_no_data_warn(temp_result, False)):
                api_base.handle_error(temp_result, '__FXBarRTIntraDayGet')
            if temp_result[:2] != '-1':
                csvString.append(temp_result if len(csvString) == 0 else temp_result[temp_result.find('\n')+1:])
        csvString = ''.join(csvString)
    
    if len(csvString) == 0:
        if 'field' not in locals() or len(field) == 0:
            field = [u'dataDate', u'currencyPair', u'barTime', u'openBid', u'highBid', u'lowBid', u'closeBid', u'openAsk', u'highAsk', u'lowAsk', u'closeAsk', u'totalTicks']
        if hasattr(field, '__iter__') and not isinstance(field, str):
            csvString = ','.join(field) + '\n'
        else:
            csvString = field + '\n'
    if pandas != "1":
        put_data_in_cache(func_name, cache_key, csvString)
        return csvString
    try:
        myIO = StringIO(csvString)
        pdFrame = pd.read_csv(myIO, dtype = {'currencyPair': 'str','barTime': 'str'},  )
        put_data_in_cache(func_name, cache_key, pdFrame)
        return pdFrame
    except Exception as e:
        raise e
    finally:
        myIO.close()

def __FXBarHistOneDayGet(currencyPair, tradeDate, field = "", pandas = "1"):
    """
    获取外汇交易市场历史某一日的分钟线数据
    
    :param currencyPair: 货币对，包含EUR|USD、USD|JPY、GBP|USD、USD|CHF、EUR|CHF、AUD|USD、AUD|CHF、USD|CAD、NZD|USD、EUR|GBP、USD|CNH、EUR|JPY、GBP|JPY、EUR|AUD、AUD|JPY、USD|HKD、GBP|CAD、GBP|AUD、NZD|CAD、AUD|NZD
    :param tradeDate: 交易日期，格式为YYYYMMDD
    :param field: 所需字段,可以是列表,可空
    :param pandas: 1表示返回 pandas data frame，0表示返回csv,可空
    :return: :raise e: API查询的结果，是CSV或者被转成pandas data frame；若查询API失败，返回空data frame； 若解析失败，则抛出异常
    """
        
    pretty_traceback()
    frame = inspect.currentframe()
    func_name, cache_key = get_cache_key(frame)
    cache_result = get_data_from_cache(func_name, cache_key)
    if cache_result is not None:
        return cache_result
    split_index = None
    split_param = None
    httpClient = api_base.__getConn__()    
    requestString = []
    requestString.append('/api/foreignMarket/getFXBarHistOneDay.csv?ispandas=1&') 
    if not isinstance(currencyPair, str) and not isinstance(currencyPair, unicode):
        currencyPair = str(currencyPair)

    requestString.append("currencyPair=%s"%(currencyPair))
    if not isinstance(tradeDate, str) and not isinstance(tradeDate, unicode):
        tradeDate = str(tradeDate)

    requestString.append("&tradeDate=%s"%(tradeDate))
    requestString.append("&field=")
    if hasattr(field,'__iter__') and not isinstance(field, str):
        if len(field) > 100 and split_param is None:
            split_index = len(requestString)
            split_param = field
            requestString.append(None)
        else:
            requestString.append(','.join(field))
    else:
        requestString.append(field)
    if split_param is None:
        csvString = api_base.__getCSV__(''.join(requestString), httpClient, gw=True)
        if csvString is None or len(csvString) == 0 or (csvString[0] == '-' and not api_base.is_no_data_warn(csvString, False)) or csvString[0] == '{':
            api_base.handle_error(csvString, '__FXBarHistOneDayGet')
        elif csvString[:2] == '-1':
            csvString = ''
    else:
        p_list = api_base.splist(split_param, 100)
        csvString = []
        for index, item in enumerate(p_list):
            requestString[split_index] = ','.join(item)
            temp_result = api_base.__getCSV__(''.join(requestString), httpClient, gw=True)
            if temp_result is None or len(temp_result) == 0 or temp_result[0] == '{' or (temp_result[0] == '-' and not api_base.is_no_data_warn(temp_result, False)):
                api_base.handle_error(temp_result, '__FXBarHistOneDayGet')
            if temp_result[:2] != '-1':
                csvString.append(temp_result if len(csvString) == 0 else temp_result[temp_result.find('\n')+1:])
        csvString = ''.join(csvString)
    
    if len(csvString) == 0:
        if 'field' not in locals() or len(field) == 0:
            field = [u'currencyPair', u'tradeDate', u'barTime', u'openBid', u'highBid', u'lowBid', u'closeBid', u'openAsk', u'highAsk', u'lowAsk', u'closeAsk', u'totalTick']
        if hasattr(field, '__iter__') and not isinstance(field, str):
            csvString = ','.join(field) + '\n'
        else:
            csvString = field + '\n'
    if pandas != "1":
        put_data_in_cache(func_name, cache_key, csvString)
        return csvString
    try:
        myIO = StringIO(csvString)
        pdFrame = pd.read_csv(myIO, dtype = {'currencyPair': 'str','barTime': 'str'},  )
        put_data_in_cache(func_name, cache_key, pdFrame)
        return pdFrame
    except Exception as e:
        raise e
    finally:
        myIO.close()

def __FXTickHistOneDayGet(currencyPair, tradeDate, field = "", pandas = "1"):
    """
    获取外汇交易市场历史某一日的Tick数据
    
    :param currencyPair: 货币对，包含EUR|USD、USD|JPY、GBP|USD、USD|CHF、EUR|CHF、AUD|USD、AUD|CHF、USD|CAD、NZD|USD、EUR|GBP、USD|CNH、EUR|JPY、GBP|JPY、EUR|AUD、AUD|JPY、USD|HKD、GBP|CAD、GBP|AUD、NZD|CAD、AUD|NZD
    :param tradeDate: 更新日期，经格林威治时间转为北京时间
    :param field: 所需字段,可以是列表,可空
    :param pandas: 1表示返回 pandas data frame，0表示返回csv,可空
    :return: :raise e: API查询的结果，是CSV或者被转成pandas data frame；若查询API失败，返回空data frame； 若解析失败，则抛出异常
    """
        
    pretty_traceback()
    frame = inspect.currentframe()
    func_name, cache_key = get_cache_key(frame)
    cache_result = get_data_from_cache(func_name, cache_key)
    if cache_result is not None:
        return cache_result
    split_index = None
    split_param = None
    httpClient = api_base.__getConn__()    
    requestString = []
    requestString.append('/api/foreignMarket/getFXTickHistOneDay.csv?ispandas=1&') 
    if not isinstance(currencyPair, str) and not isinstance(currencyPair, unicode):
        currencyPair = str(currencyPair)

    requestString.append("currencyPair=%s"%(currencyPair))
    if not isinstance(tradeDate, str) and not isinstance(tradeDate, unicode):
        tradeDate = str(tradeDate)

    requestString.append("&tradeDate=%s"%(tradeDate))
    requestString.append("&field=")
    if hasattr(field,'__iter__') and not isinstance(field, str):
        if len(field) > 100 and split_param is None:
            split_index = len(requestString)
            split_param = field
            requestString.append(None)
        else:
            requestString.append(','.join(field))
    else:
        requestString.append(field)
    if split_param is None:
        csvString = api_base.__getCSV__(''.join(requestString), httpClient, gw=True)
        if csvString is None or len(csvString) == 0 or (csvString[0] == '-' and not api_base.is_no_data_warn(csvString, False)) or csvString[0] == '{':
            api_base.handle_error(csvString, '__FXTickHistOneDayGet')
        elif csvString[:2] == '-1':
            csvString = ''
    else:
        p_list = api_base.splist(split_param, 100)
        csvString = []
        for index, item in enumerate(p_list):
            requestString[split_index] = ','.join(item)
            temp_result = api_base.__getCSV__(''.join(requestString), httpClient, gw=True)
            if temp_result is None or len(temp_result) == 0 or temp_result[0] == '{' or (temp_result[0] == '-' and not api_base.is_no_data_warn(temp_result, False)):
                api_base.handle_error(temp_result, '__FXTickHistOneDayGet')
            if temp_result[:2] != '-1':
                csvString.append(temp_result if len(csvString) == 0 else temp_result[temp_result.find('\n')+1:])
        csvString = ''.join(csvString)
    
    if len(csvString) == 0:
        if 'field' not in locals() or len(field) == 0:
            field = [u'offerID', u'currencyPair', u'quoteID', u'bid', u'ask', u'Low', u'High', u'volume', u'updateDate', u'updateTime', u'sellInterest', u'buyInterest', u'contractCurrency', u'Digits', u'pointSize', u'contractMultiplier', u'tradingStatus', u'valueDate']
        if hasattr(field, '__iter__') and not isinstance(field, str):
            csvString = ','.join(field) + '\n'
        else:
            csvString = field + '\n'
    if pandas != "1":
        put_data_in_cache(func_name, cache_key, csvString)
        return csvString
    try:
        myIO = StringIO(csvString)
        pdFrame = pd.read_csv(myIO, dtype = {'currencyPair': 'str','quoteID': 'str','contractCurrency': 'str','tradingStatus': 'str'},  )
        put_data_in_cache(func_name, cache_key, pdFrame)
        return pdFrame
    except Exception as e:
        raise e
    finally:
        myIO.close()




#coding=utf-8
global_cache_map = {}

class Settings(object):
    '''
    DataAPI的设置
    '''
    def __init__(self):
        self._cache_enabled = False
        self.pretty_traceback_enabled = True
    @property
    def cache_enabled(self):
        """DataAPI cache功能是否开启"""
        return self._cache_enabled

    @cache_enabled.setter
    def cache_enabled(self, value):
        if not value:
            global_cache_map.clear()
        self._cache_enabled = value


# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os
import traceback
import requests

from . import utils
from .trading import LiveTrading
from .user_factor import UserFactor
from .notebook import Notebook
from .data import Data
from .authorize import Authorize

### python 3.x unicode
try:
    unicode
except:
    unicode = str


session = requests.Session()

class Client(Authorize):
    """优矿
    """
    def __init__(self, username='', password='', token='', account_file='', data_target_path='', notebook_target_path=''):

        Authorize.__init__(self, username, password, token, account_file, session)
        
        ### Authorize
        self._authorize()

        if self._isvalid:
            ### Trading
            self.trading = LiveTrading(self.session)

            ### notebook，data，library
            self.data = Data(self.session, data_target_path)
            self.notebook = Notebook(self.session, notebook_target_path)

            ## factor
            self.factor = UserFactor(self.session)
    


def sdk_version(auto_update=False):
    """查询优矿专业版 SDK 版本信息

    参数：
        auto_update: 设置为 True 时可以自动更新，设置为 False 时用户可以根据函数打印的下载链接自行下载安装。
    """
    utils.version_check(auto_update)
